# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections import OrderedDict
from collections.abc import ItemsView, Iterable, KeysView, Mapping, ValuesView
from enum import Enum
from typing import Any, NamedTuple, Union

from braket.circuits.compiler_directive import CompilerDirective
from braket.circuits.gate import Gate
from braket.circuits.instruction import Instruction
from braket.circuits.measure import Measure
from braket.circuits.noise import Noise
from braket.registers.qubit import Qubit
from braket.registers.qubit_set import QubitSet


class MomentType(str, Enum):
    """The type of moments.
    GATE: a gate
    NOISE: a noise channel added directly to the circuit
    GATE_NOISE: a gate-based noise channel
    INITIALIZATION_NOISE: a initialization noise channel
    READOUT_NOISE: a readout noise channel
    COMPILER_DIRECTIVE: an instruction to the compiler, external to the quantum program itself
    MEASURE: a measurement
    """

    GATE = "gate"
    NOISE = "noise"
    GATE_NOISE = "gate_noise"
    INITIALIZATION_NOISE = "initialization_noise"
    READOUT_NOISE = "readout_noise"
    COMPILER_DIRECTIVE = "compiler_directive"
    GLOBAL_PHASE = "global_phase"
    MEASURE = "measure"


class MomentsKey(NamedTuple):
    """Key of the Moments mapping.

    Args:
        time: moment
        qubits: qubit set
        moment_type: The type of the moment
        noise_index: the number of noise channels at the same moment. For gates, this is the
            number of gate_noise channels associated with that gate. For all other noise
            types, noise_index starts from 0; but for gate noise, it starts from 1.
    """

    time: int
    qubits: QubitSet
    moment_type: MomentType
    noise_index: int
    subindex: int = 0


class Moments(Mapping[MomentsKey, Instruction]):
    r"""An ordered mapping of `MomentsKey` or `NoiseMomentsKey` to `Instruction`. The
    core data structure that contains instructions, ordering they are inserted in, and
    time slices when they occur. `Moments` implements `Mapping` and functions the same as
    a read-only dictionary. It is mutable only through the `add()` method.

    This data structure is useful to determine a dependency of instructions, such as
    printing or optimizing circuit structure, before sending it to a quantum
    device. The original insertion order is preserved and can be retrieved via the `values()`
    method.

    Args:
        instructions (Iterable[Instruction] | None): Instructions to initialize self.
            Default = None.

    Examples:
        >>> moments = Moments()
        >>> moments.add([Instruction(Gate.H(), 0), Instruction(Gate.CNot(), [0, 1])])
        >>> moments.add([Instruction(Gate.H(), 0), Instruction(Gate.H(), 1)])
        >>> for i, item in enumerate(moments.items()):
        ...     print(f"Item {i}")
        ...     print(f"\\tKey: {item[0]}")
        ...     print(f"\\tValue: {item[1]}")
        ...
        Item 0
            Key: MomentsKey(time=0, qubits=QubitSet([Qubit(0)]))
            Value: Instruction('operator': H, 'target': QubitSet([Qubit(0)]))
        Item 1
            Key: MomentsKey(time=1, qubits=QubitSet([Qubit(0), Qubit(1)]))
            Value: Instruction('operator': CNOT, 'target': QubitSet([Qubit(0), Qubit(1)]))
        Item 2
            Key: MomentsKey(time=2, qubits=QubitSet([Qubit(0)]))
            Value: Instruction('operator': H, 'target': QubitSet([Qubit(0)]))
        Item 3
            Key: MomentsKey(time=2, qubits=QubitSet([Qubit(1)]))
            Value: Instruction('operator': H, 'target': QubitSet([Qubit(1)]))
    """

    def __init__(self, instructions: Iterable[Instruction] | None = None):
        self._moments: OrderedDict[MomentsKey, Instruction] = OrderedDict()
        self._max_times: dict[Qubit, int] = {}
        self._qubits = QubitSet()
        self._depth = 0
        self._time_all_qubits = -1
        self._number_gphase_in_current_moment = 0

        self.add(instructions or [])

    @property
    def depth(self) -> int:
        """int: Get the depth (number of slices) of self."""
        return self._depth

    @property
    def qubit_count(self) -> int:
        """int: Get the number of qubits used across all of the instructions."""
        return len(self._qubits)

    @property
    def qubits(self) -> QubitSet:
        """QubitSet: Get the qubits used across all of the instructions. The order of qubits is
        based on the order in which the instructions were added.

        Note:
            Don't mutate this object, any changes may impact the behavior of this class and / or
            consumers. If you need to mutate this, then copy it via `QubitSet(moments.qubits())`.
        """
        return self._qubits

    def time_slices(self) -> dict[int, list[Instruction]]:
        """Get instructions keyed by time.

        Returns:
            dict[int, list[Instruction]]: Key is the time and value is a list of instructions that
            occur at that moment in time. The order of instructions is in no particular order.

        Note:
            This is a computed result over self and can be freely mutated. This is re-computed with
            every call, with a computational runtime O(N) where N is the number
            of instructions in self.
        """
        time_slices = {}
        self.sort_moments()
        for key, instruction in self._moments.items():
            instructions = time_slices.get(key.time, [])
            instructions.append(instruction)
            time_slices[key.time] = instructions

        return time_slices

    def add(
        self, instructions: Union[Iterable[Instruction], Instruction], noise_index: int = 0
    ) -> None:
        """Add one or more instructions to self.

        Args:
            instructions (Union[Iterable[Instruction], Instruction]): Instructions to add to self.
                The instruction is added to the max time slice in which the instruction fits.
            noise_index (int): the number of noise channels at the same moment. For gates, this
                is the number of gate_noise channels associated with that gate. For all other noise
                types, noise_index starts from 0; but for gate noise, it starts from 1.
        """
        if isinstance(instructions, Instruction):
            instructions = [instructions]
        for instruction in instructions:
            self._add(instruction, noise_index)

    def _add(self, instruction: Instruction, noise_index: int = 0) -> None:
        operator = instruction.operator
        if isinstance(operator, CompilerDirective):
            time = self._update_qubit_times(self._qubits)
            self._moments[MomentsKey(time, None, MomentType.COMPILER_DIRECTIVE, 0)] = instruction
            self._depth = time + 1
            self._time_all_qubits = time
        elif isinstance(operator, Noise):
            self.add_noise(instruction)
        elif isinstance(operator, Gate) and operator.name == "GPhase":
            time = self._get_qubit_times(self._max_times.keys()) + 1
            self._number_gphase_in_current_moment += 1
            key = MomentsKey(
                time,
                QubitSet([]),
                MomentType.GLOBAL_PHASE,
                0,
                self._number_gphase_in_current_moment,
            )
            self._moments[key] = instruction
        elif isinstance(operator, Measure):
            qubit_range = instruction.target.union(instruction.control)
            time = self._get_qubit_times(self._max_times.keys()) + 1
            self._moments[MomentsKey(time, qubit_range, MomentType.MEASURE, noise_index)] = (
                instruction
            )
            self._qubits.update(qubit_range)
            self._depth = max(self._depth, time + 1)
        else:
            qubit_range = instruction.target.union(instruction.control)
            time = self._update_qubit_times(qubit_range)
            self._moments[MomentsKey(time, qubit_range, MomentType.GATE, noise_index)] = instruction
            self._qubits.update(qubit_range)
            self._depth = max(self._depth, time + 1)

    def _get_qubit_times(self, qubits: QubitSet) -> int:
        return max([self._max_time_for_qubit(qubit) for qubit in qubits] + [self._time_all_qubits])

    def _update_qubit_times(self, qubits: QubitSet) -> int:
        time = self._get_qubit_times(qubits) + 1
        # Update time for all specified qubits
        for qubit in qubits:
            self._max_times[qubit] = time
        self._number_gphase_in_current_moment = 0
        return time

    def add_noise(
        self, instruction: Instruction, input_type: str = "noise", noise_index: int = 0
    ) -> None:
        """Adds noise to a moment.

        Args:
            instruction (Instruction): Instruction to add.
            input_type (str): One of MomentType.
            noise_index (int): The number of noise channels at the same moment. For gates, this
                is the number of gate_noise channels associated with that gate. For all other noise
                types, noise_index starts from 0; but for gate noise, it starts from 1.
        """
        qubit_range = instruction.target
        time = max(0, *[self._max_time_for_qubit(qubit) for qubit in qubit_range])
        if input_type == MomentType.INITIALIZATION_NOISE:
            time = 0

        while MomentsKey(time, qubit_range, input_type, noise_index) in self._moments:
            noise_index = noise_index + 1

        self._moments[MomentsKey(time, qubit_range, input_type, noise_index)] = instruction
        self._qubits.update(qubit_range)

    def sort_moments(self) -> None:
        """Make the disordered moments in order.

        1. Make the readout noise in the end
        2. Make the initialization noise at the beginning
        """
        # key for NOISE, GATE and GATE_NOISE
        key_noise = []
        # key for INITIALIZATION_NOISE
        key_initialization_noise = []
        # key for READOUT_NOISE
        key_readout_noise = []
        moment_copy = OrderedDict()
        sorted_moment = OrderedDict()

        for key, instruction in self._moments.items():
            moment_copy[key] = instruction
            if key.moment_type == MomentType.READOUT_NOISE:
                key_readout_noise.append(key)
            elif key.moment_type == MomentType.INITIALIZATION_NOISE:
                key_initialization_noise.append(key)
            else:
                key_noise.append(key)

        for key in key_initialization_noise:
            sorted_moment[key] = moment_copy[key]
        for key in key_noise:
            sorted_moment[key] = moment_copy[key]
        # find the max time in the circuit and make it the time for readout noise
        max_time = max(self._depth - 1, 0)

        for key in key_readout_noise:
            sorted_moment[
                MomentsKey(max_time, key.qubits, MomentType.READOUT_NOISE, key.noise_index)
            ] = moment_copy[key]

        self._moments = sorted_moment

    def _max_time_for_qubit(self, qubit: Qubit) -> int:
        # -1 if qubit is unoccupied because the first instruction will have an index of 0
        return self._max_times.get(qubit, -1)

    #
    # Implement abstract methods, default to calling selfs underlying dictionary
    #

    def keys(self) -> KeysView[MomentsKey]:
        """Return a view of self's keys."""
        return self._moments.keys()

    def items(self) -> ItemsView[MomentsKey, Instruction]:
        """Return a view of self's (key, instruction)."""
        return self._moments.items()

    def values(self) -> ValuesView[Instruction]:
        """Return a view of self's instructions.

        Returns:
            ValuesView[Instruction]: The (in-order) instructions.
        """
        self.sort_moments()
        return self._moments.values()

    def get(self, key: MomentsKey, default: Any | None = None) -> Instruction:
        """Get the instruction in self by key.

        Args:
            key (MomentsKey): Key of the instruction to fetch.
            default (Any | None): Value to return if `key` is not in `moments`. Default = `None`.

        Returns:
            Instruction: `moments[key]` if `key` in `moments`, else `default` is returned.
        """
        return self._moments.get(key, default)

    def __getitem__(self, key: MomentsKey):
        return self._moments.__getitem__(key)

    def __iter__(self):
        return self._moments.__iter__()

    def __len__(self):
        return self._moments.__len__()

    def __contains__(self, item: MomentsKey):
        return self._moments.__contains__(item)

    def __eq__(self, other: Moments):
        if isinstance(other, Moments):
            return self._moments == other._moments
        return NotImplemented

    def __ne__(self, other: Moments):
        result = self.__eq__(other)
        if result is not NotImplemented:
            return not result
        return NotImplemented

    def __repr__(self):
        return self._moments.__repr__()

    def __str__(self):
        return self._moments.__str__()



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.parametric import FreeParameter


@pytest.fixture
def free_parameter():
    return FreeParameter("theta")


def test_bad_input():
    with pytest.raises(ValueError, match="FreeParameter names must be non empty"):
        FreeParameter("")
    with pytest.raises(TypeError, match="FreeParameter names must be strings"):
        FreeParameter(6)
    with pytest.raises(
        ValueError, match="FreeParameter names must start with a letter or an underscore"
    ):
        FreeParameter(".2")


def test_is_free_param(free_parameter):
    assert isinstance(free_parameter, FreeParameter)


def test_equality():
    param_1 = FreeParameter("theta")
    param_2 = FreeParameter("theta")
    other_param = FreeParameter("phi")
    non_param = "non circuit"

    assert param_1 == param_2
    assert param_1 is not param_2
    assert param_1 != other_param
    assert param_1 != non_param


def test_str(free_parameter):
    expected = "theta"
    assert str(free_parameter) == expected


def test_hash(free_parameter):
    assert hash(free_parameter) == hash(tuple(free_parameter.name))


def test_rep(free_parameter):
    assert repr(free_parameter) == free_parameter.name


def test_sub_successful(free_parameter):
    assert free_parameter.subs({"theta": 1}) == 1


def test_sub_wrong_param(free_parameter):
    assert free_parameter.subs({"alpha": 1}) == FreeParameter("theta")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numpy as np
import pytest

from braket.circuits import Operator, QuantumOperator


class _DummyQuantumOperator(QuantumOperator):
    @staticmethod
    def fixed_qubit_count():
        return 2


@pytest.fixture
def ascii_symbols():
    return ["foo"]


@pytest.fixture
def quantum_operator(ascii_symbols):
    return QuantumOperator(qubit_count=1, ascii_symbols=ascii_symbols)


def test_is_operator(quantum_operator):
    assert isinstance(quantum_operator, Operator)


def test_ascii_symbols(quantum_operator, ascii_symbols):
    assert quantum_operator.ascii_symbols == tuple(ascii_symbols)


def test_fixed_qubit_count_implemented():
    operator = _DummyQuantumOperator(qubit_count=None, ascii_symbols=["foo", "bar"])
    assert operator.qubit_count == _DummyQuantumOperator.fixed_qubit_count()


def test_qubit_count_fixed_qubit_count_unequal():
    with pytest.raises(ValueError):
        _DummyQuantumOperator(qubit_count=1, ascii_symbols=["foo", "bar"])


def test_qubit_count_not_int():
    with pytest.raises(TypeError):
        QuantumOperator(qubit_count="hello", ascii_symbols=[])


def test_qubit_count_lt_one():
    with pytest.raises(ValueError):
        QuantumOperator(qubit_count=0, ascii_symbols=[])


def test_none_ascii():
    with pytest.raises(ValueError):
        QuantumOperator(qubit_count=1, ascii_symbols=None)


def test_mismatch_length_ascii():
    with pytest.raises(ValueError):
        QuantumOperator(qubit_count=1, ascii_symbols=["foo", "bar"])


def test_name(quantum_operator):
    expected = quantum_operator.__class__.__name__
    assert quantum_operator.name == expected


def test_getters():
    qubit_count = 2
    ascii_symbols = ("foo", "bar")
    quantum_operator = QuantumOperator(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

    assert quantum_operator.qubit_count == qubit_count
    assert quantum_operator.ascii_symbols == ascii_symbols


def test_qubit_count_setter(quantum_operator):
    with pytest.raises(AttributeError):
        quantum_operator.qubit_count = 10


def test_ascii_symbols_setter(quantum_operator):
    with pytest.raises(AttributeError):
        quantum_operator.ascii_symbols = ["foo", "bar"]


def test_name_setter(quantum_operator):
    with pytest.raises(AttributeError):
        quantum_operator.name = "hi"


def test_to_ir_not_implemented_by_default(quantum_operator):
    with pytest.raises(NotImplementedError):
        quantum_operator.to_ir(None)


def test_to_matrix_not_implemented_by_default(quantum_operator):
    with pytest.raises(NotImplementedError):
        quantum_operator.to_matrix(None)


def test_matrix_equivalence():
    class _Foo1(QuantumOperator):
        def __init__(self):
            super().__init__(qubit_count=1, ascii_symbols=["foo"])

        def to_matrix(self):
            return np.array([[1.0, 0.0], [0.0, 1.0j]])

    class _Foo2(QuantumOperator):
        def __init__(self):
            super().__init__(qubit_count=1, ascii_symbols=["foo"])

        def to_matrix(self):
            return np.array([[1.0, 0.0], [1.0, 0.0]])

    quantum_operator1 = _Foo1()
    quantum_operator2 = _Foo1()
    quantum_operator3 = _Foo2()
    assert quantum_operator1.matrix_equivalence(quantum_operator2)
    assert not quantum_operator1.matrix_equivalence(quantum_operator3)


def test_matrix_equivalence_non_quantum_operator():
    quantum_operator1 = QuantumOperator(qubit_count=1, ascii_symbols=["foo"])
    assert not quantum_operator1.matrix_equivalence(1)


def test_str(quantum_operator):
    expected = "{}('qubit_count': {})".format(quantum_operator.name, quantum_operator.qubit_count)
    assert str(quantum_operator) == expected



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from abc import abstractmethod
from typing import Optional, Union

from braket.circuits.instruction import Instruction
from braket.circuits.noise_model.criteria import Criteria
from braket.registers.qubit_set import QubitSetInput


class CircuitInstructionCriteria(Criteria):
    """Criteria that implement these methods may be used to determine gate noise."""

    @abstractmethod
    def instruction_matches(self, instruction: Instruction) -> bool:
        """Returns True if an Instruction matches the criteria.

        Args:
            instruction (Instruction): An Instruction to match.

        Raises:
            NotImplementedError: Not implemented.

        Returns:
            bool: True if an Instruction matches the criteria.
        """
        raise NotImplementedError

    @staticmethod
    def _check_target_in_qubits(
        qubits: Optional[set[Union[int, tuple[int]]]], target: QubitSetInput
    ) -> bool:
        """Returns true if the given targets of an instruction match the given qubit input set.

        Args:
            qubits (Optional[set[Union[int, tuple[int]]]]): The qubits provided to the criteria.
            target (QubitSetInput): Targets of an instruction.

        Returns:
            bool: True if the provided target should be matched by the given qubits.
        """
        if qubits is None:
            return True
        target = [int(item) for item in target]
        if len(target) == 1:
            return target[0] in qubits
        return tuple(target) in qubits



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from enum import Enum

import braket.ir.annealing as ir


class ProblemType(str, Enum):
    """The type of annealing problem.

    QUBO: Quadratic Unconstrained Binary Optimization, with values 1 and 0

    ISING: Ising model, with values +/-1
    """

    QUBO = "QUBO"
    ISING = "ISING"


class Problem:
    """Represents an annealing problem."""

    def __init__(
        self,
        problem_type: ProblemType,
        linear: dict[int, float] | None = None,
        quadratic: dict[tuple[int, int], float] | None = None,
    ):
        """Initialzes a `Problem`.

        Args:
            problem_type (ProblemType): The type of annealing problem
            linear (dict[int, float] | None): The linear terms of this problem,
                as a map of variable to coefficient
            quadratic (dict[tuple[int, int], float] | None): The quadratic terms of this problem,
                as a map of variables to coefficient

        Examples:
            >>> problem = Problem(
            >>>     ProblemType.ISING,
            >>>     linear={1: 3.14},
            >>>     quadratic={(1, 2): 10.08},
            >>> )
            >>> problem.add_linear_term(2, 1.618).add_quadratic_term((3, 4), 1337)
        """
        self._problem_type = problem_type
        self._linear = linear or {}
        self._quadratic = quadratic or {}

    @property
    def problem_type(self) -> ProblemType:
        """The type of annealing problem.

        Returns:
            ProblemType: The type of annealing problem
        """
        return self._problem_type

    @property
    def linear(self) -> dict[int, float]:
        """The linear terms of this problem.

        Returns:
            dict[int, float]: The linear terms of this problem, as a map of variable to coefficient
        """
        return self._linear

    @property
    def quadratic(self) -> dict[tuple[int, int], float]:
        """The quadratic terms of this problem.

        Returns:
            dict[tuple[int, int], float]: The quadratic terms of this problem,
            as a map of variables to coefficient
        """
        return self._quadratic

    def add_linear_term(self, term: int, coefficient: float) -> Problem:
        """Adds a linear term to the problem.

        Args:
            term (int): The variable of the linear term
            coefficient (float): The coefficient of the linear term

        Returns:
            Problem: This problem object
        """
        self._linear[term] = coefficient
        return self

    def add_linear_terms(self, coefficients: dict[int, float]) -> Problem:
        """Adds linear terms to the problem.

        Args:
            coefficients (dict[int, float]): A map of variable to coefficient

        Returns:
            Problem: This problem object
        """
        self._linear.update(coefficients)
        return self

    def add_quadratic_term(self, term: tuple[int, int], coefficient: float) -> Problem:
        """Adds a quadratic term to the problem.

        Args:
            term (tuple[int, int]): The variables of the quadratic term
            coefficient (float): The coefficient of the quadratic term

        Returns:
            Problem: This problem object
        """
        self._quadratic[term] = coefficient
        return self

    def add_quadratic_terms(self, coefficients: dict[tuple[int, int], float]) -> Problem:
        """Adds quadratic terms to the problem.

        Args:
            coefficients (dict[tuple[int, int], float]): A map of variables to coefficient

        Returns:
            Problem: This problem object
        """
        self._quadratic.update(coefficients)
        return self

    def to_ir(self) -> Problem:
        """Converts this problem into IR representation.

        Returns:
            Problem: IR representation of this problem object
        """
        return ir.Problem(
            type=ir.ProblemType[self._problem_type.value],
            linear=self._linear,
            quadratic={
                ",".join((str(q1), str(q2))): self._quadratic[(q1, q2)]
                for q1, q2 in self._quadratic
            },
        )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import uuid

import numpy as np
import pytest

from braket.task_result import TaskMetadata
from braket.tasks import GateModelQuantumTaskResult
from braket.tasks.local_quantum_task import LocalQuantumTask

RESULT = GateModelQuantumTaskResult(
    task_metadata=TaskMetadata(**{"id": str(uuid.uuid4()), "deviceId": "default", "shots": 100}),
    additional_metadata=None,
    measurements=np.array([[0, 1], [1, 0]]),
    measured_qubits=[0, 1],
    result_types=None,
    values=None,
)

TASK = LocalQuantumTask(RESULT)


def test_id():
    # Task ID is valid UUID
    uuid.UUID(TASK.id)


def test_state():
    assert TASK.state() == "COMPLETED"


def test_result():
    assert RESULT.task_metadata.id == TASK.id
    assert TASK.result() == RESULT


@pytest.mark.xfail(raises=NotImplementedError)
def test_cancel():
    TASK.cancel()


@pytest.mark.xfail(raises=NotImplementedError)
def test_async():
    TASK.async_result()


def test_str():
    expected = "LocalQuantumTask('id':{})".format(TASK.id)
    assert str(TASK) == expected



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from decimal import Decimal
from unittest.mock import Mock

import pytest

from braket.ahs.discretization_types import DiscretizationError
from braket.ahs.field import Field
from braket.ahs.pattern import Pattern
from braket.timings.time_series import TimeSeries


@pytest.fixture
def default_time_series():
    default_values = [(2700, 25.1327), (300, 25.1327), (600, 15.1327), (Decimal(0.3), Decimal(0.4))]
    time_series = TimeSeries()
    for value in default_values:
        time_series.put(value[0], value[1])
    return time_series


@pytest.fixture
def default_pattern():
    return Pattern(series=[0, 0.1, 1, 0.5, 0.2, 0.001, 1e-10])


@pytest.fixture
def default_field(default_time_series, default_pattern):
    return Field(time_series=default_time_series, pattern=default_pattern)


@pytest.fixture
def default_uniform_field(default_time_series):
    return Field(time_series=default_time_series)


def test_create():
    mock0 = Mock()
    mock1 = Mock()
    field = Field(time_series=mock0, pattern=mock1)
    assert mock0 == field.time_series
    assert mock1 == field.pattern


@pytest.mark.parametrize(
    "time_res, value_res, pattern_res",
    [
        (Decimal("0.1"), Decimal("10"), Decimal("0.5")),
        (Decimal("10"), Decimal("20"), Decimal("0.1")),
        (Decimal("100"), Decimal("0.1"), Decimal("1")),
    ],
)
def test_discretize(
    default_time_series, default_pattern, default_field, time_res, value_res, pattern_res
):
    expected = Field(
        time_series=default_time_series.discretize(time_res, value_res),
        pattern=default_pattern.discretize(pattern_res),
    )
    actual = default_field.discretize(time_res, value_res, pattern_res)
    assert expected.pattern.series == actual.pattern.series
    assert expected.time_series.times() == actual.time_series.times()
    assert expected.time_series.values() == actual.time_series.values()


@pytest.mark.parametrize(
    "time_res, value_res, pattern_res",
    [
        (Decimal("0.1"), Decimal("10"), Decimal("0.5")),
        (Decimal("10"), Decimal("20"), None),
        (Decimal("100"), Decimal("0.1"), Decimal("1")),
    ],
)
def test_uniform_field(
    default_time_series, default_uniform_field, time_res, value_res, pattern_res
):
    expected = Field(time_series=default_time_series.discretize(time_res, value_res))
    actual = default_uniform_field.discretize(time_res, value_res, pattern_res)
    assert (
        (expected.pattern is None) and (actual.pattern is None)
    ) or expected.pattern.series == actual.pattern.series
    assert expected.time_series.times() == actual.time_series.times()
    assert expected.time_series.values() == actual.time_series.values()


@pytest.mark.parametrize(
    "time_res, value_res, pattern_res",
    [
        (Decimal("10"), Decimal("20"), None),
    ],
)
@pytest.mark.xfail(raises=DiscretizationError)
def test_invalid_pattern_res(default_field, time_res, value_res, pattern_res):
    default_field.discretize(time_res, value_res, pattern_res)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
from pathlib import Path
from unittest.mock import Mock, mock_open, patch

import pytest

from braket.jobs.local.local_job_container_setup import setup_container


@pytest.fixture
def aws_session():
    _aws_session = Mock()
    _aws_session.boto_session.get_credentials.return_value.access_key = "Test Access Key"
    _aws_session.boto_session.get_credentials.return_value.secret_key = "Test Secret Key"
    _aws_session.boto_session.get_credentials.return_value.token = None
    _aws_session.region = "Test Region"
    _aws_session.list_keys.side_effect = lambda bucket, prefix: [
        key
        for key in [
            "input-dir/",
            "input-dir/file-1.txt",
            "input-dir/file-2.txt",
        ]
        if key.startswith(prefix)
    ]
    return _aws_session


@pytest.fixture
def container():
    _container = Mock()
    return _container


@pytest.fixture
def creation_kwargs():
    return {
        "algorithmSpecification": {
            "scriptModeConfig": {
                "entryPoint": "my_file:start_here",
                "s3Uri": "s3://amazon-braket-jobs/job-path/my_file.py",
            }
        },
        "checkpointConfig": {
            "localPath": "/opt/omega/checkpoints",
            "s3Uri": "s3://amazon-braket-jobs/job-path/checkpoints",
        },
        "outputDataConfig": {"s3Path": "s3://test_bucket/test_location/"},
        "deviceConfig": {"device": "test device ARN"},
        "jobName": "Test-Job-Name",
        "roleArn": "arn:aws:iam::875981177017:role/AmazonBraketJobRole",
    }


@pytest.fixture
def compressed_script_mode_config():
    return {
        "scriptModeConfig": {
            "entryPoint": "my_file:start_here",
            "s3Uri": "s3://amazon-braket-jobs/job-path/my_archive.gzip",
            "compressionType": "gzip",
        }
    }


@pytest.fixture
def expected_envs():
    return {
        "AMZN_BRAKET_CHECKPOINT_DIR": "/opt/omega/checkpoints",
        "AMZN_BRAKET_DEVICE_ARN": "test device ARN",
        "AMZN_BRAKET_JOB_NAME": "Test-Job-Name",
        "AMZN_BRAKET_JOB_RESULTS_DIR": "/opt/braket/model",
        "AMZN_BRAKET_JOB_RESULTS_S3_PATH": "test_location/Test-Job-Name/output",
        "AMZN_BRAKET_OUT_S3_BUCKET": "test_bucket",
        "AMZN_BRAKET_SCRIPT_ENTRY_POINT": "my_file:start_here",
        "AMZN_BRAKET_SCRIPT_S3_URI": "s3://amazon-braket-jobs/job-path/my_file.py",
        "AMZN_BRAKET_TASK_RESULTS_S3_URI": "s3://test_bucket/jobs/Test-Job-Name/tasks",
        "AWS_ACCESS_KEY_ID": "Test Access Key",
        "AWS_DEFAULT_REGION": "Test Region",
        "AWS_SECRET_ACCESS_KEY": "Test Secret Key",
    }


@pytest.fixture
def input_data_config():
    return [
        # s3 prefix is a single file
        {
            "channelName": "single-file",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/input-dir/file-1.txt"}},
        },
        # s3 prefix is a directory no slash
        {
            "channelName": "directory-no-slash",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/input-dir"}},
        },
        # s3 prefix is a directory with slash
        {
            "channelName": "directory-slash",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/input-dir/"}},
        },
        # s3 prefix is a prefix for a directory
        {
            "channelName": "directory-prefix",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/input"}},
        },
        # s3 prefix is a prefix for multiple files
        {
            "channelName": "files-prefix",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/input-dir/file"}},
        },
    ]


def test_basic_setup(container, aws_session, creation_kwargs, expected_envs):
    aws_session.parse_s3_uri.return_value = ["test_bucket", "test_location"]
    envs = setup_container(container, aws_session, **creation_kwargs)
    assert envs == expected_envs
    container.makedir.assert_any_call("/opt/ml/model")
    container.makedir.assert_any_call(expected_envs["AMZN_BRAKET_CHECKPOINT_DIR"])
    assert container.makedir.call_count == 2


def test_compressed_script_mode(
    container, aws_session, creation_kwargs, expected_envs, compressed_script_mode_config
):
    creation_kwargs["algorithmSpecification"] = compressed_script_mode_config
    expected_envs["AMZN_BRAKET_SCRIPT_S3_URI"] = "s3://amazon-braket-jobs/job-path/my_archive.gzip"
    expected_envs["AMZN_BRAKET_SCRIPT_COMPRESSION_TYPE"] = "gzip"
    aws_session.parse_s3_uri.return_value = ["test_bucket", "test_location"]
    envs = setup_container(container, aws_session, **creation_kwargs)
    assert envs == expected_envs
    container.makedir.assert_any_call("/opt/ml/model")
    container.makedir.assert_any_call(expected_envs["AMZN_BRAKET_CHECKPOINT_DIR"])
    assert container.makedir.call_count == 2


@patch("json.dump")
@patch("tempfile.TemporaryDirectory")
def test_hyperparameters(tempfile, json, container, aws_session, creation_kwargs, expected_envs):
    with patch("builtins.open", mock_open()):
        tempfile.return_value.__enter__.return_value = "temporaryDir"
        creation_kwargs["hyperParameters"] = {"test": "hyper"}
        expected_envs["AMZN_BRAKET_HP_FILE"] = "/opt/braket/input/config/hyperparameters.json"
        aws_session.parse_s3_uri.return_value = ["test_bucket", "test_location"]
        envs = setup_container(container, aws_session, **creation_kwargs)
        assert envs == expected_envs
        container.makedir.assert_any_call("/opt/ml/model")
        container.makedir.assert_any_call(expected_envs["AMZN_BRAKET_CHECKPOINT_DIR"])
        assert container.makedir.call_count == 2
        container.copy_to.assert_called_with(
            os.path.join("temporaryDir", "hyperparameters.json"),
            "/opt/ml/input/config/hyperparameters.json",
        )


def test_input(container, aws_session, creation_kwargs, input_data_config):
    creation_kwargs.update({"inputDataConfig": input_data_config})
    setup_container(container, aws_session, **creation_kwargs)
    download_locations = [call[0][1] for call in aws_session.download_from_s3.call_args_list]
    expected_downloads = [
        Path("single-file", "file-1.txt"),
        Path("directory-no-slash", "file-1.txt"),
        Path("directory-no-slash", "file-2.txt"),
        Path("directory-slash", "file-1.txt"),
        Path("directory-slash", "file-2.txt"),
        Path("directory-prefix", "input-dir", "file-1.txt"),
        Path("directory-prefix", "input-dir", "file-2.txt"),
        Path("files-prefix", "file-1.txt"),
        Path("files-prefix", "file-2.txt"),
    ]

    for download, expected_download in zip(download_locations, expected_downloads):
        assert download.endswith(str(expected_download))


def test_duplicate_input(container, aws_session, creation_kwargs, input_data_config):
    input_data_config.append(
        {
            # this is a duplicate channel
            "channelName": "single-file",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/irrelevant"}},
        }
    )
    creation_kwargs.update({"inputDataConfig": input_data_config})
    dupes_not_allowed = "Duplicate channel names not allowed for input data: single-file"
    with pytest.raises(ValueError, match=dupes_not_allowed):
        setup_container(container, aws_session, **creation_kwargs)


def test_no_data_input(container, aws_session, creation_kwargs, input_data_config):
    input_data_config.append(
        {
            # this channel won't match any data
            "channelName": "no-data",
            "dataSource": {"s3DataSource": {"s3Uri": "s3://input_bucket/irrelevant"}},
        }
    )
    creation_kwargs.update({"inputDataConfig": input_data_config})
    no_data_found = "No data found for channel 'no-data'"
    with pytest.raises(RuntimeError, match=no_data_found):
        setup_container(container, aws_session, **creation_kwargs)


def test_temporary_credentials(container, aws_session, creation_kwargs, expected_envs):
    aws_session.boto_session.get_credentials.return_value.token = "Test Token"
    expected_envs["AWS_SESSION_TOKEN"] = "Test Token"
    aws_session.parse_s3_uri.return_value = ["test_bucket", "test_location"]
    envs = setup_container(container, aws_session, **creation_kwargs)
    assert envs == expected_envs
    container.makedir.assert_any_call("/opt/ml/model")
    container.makedir.assert_any_call(expected_envs["AMZN_BRAKET_CHECKPOINT_DIR"])
    assert container.makedir.call_count == 2



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import datetime
import json
import logging
import os
import re
import tarfile
import tempfile
from unittest.mock import Mock, patch

import pytest
from botocore.exceptions import ClientError

from braket.aws import AwsQuantumJob, AwsSession
from braket.aws.queue_information import HybridJobQueueInfo


@pytest.fixture
def aws_session(quantum_job_arn, job_region):
    _aws_session = Mock(spec=AwsSession)
    _aws_session.create_job.return_value = quantum_job_arn
    _aws_session.default_bucket.return_value = "default-bucket-name"
    _aws_session.get_default_jobs_role.return_value = "default-role-arn"
    _aws_session.construct_s3_uri.side_effect = (
        lambda bucket, *dirs: f"s3://{bucket}/{'/'.join(dirs)}"
    )

    def fake_copy_session(region):
        _aws_session.region = region
        return _aws_session

    _aws_session.copy_session.side_effect = fake_copy_session
    _aws_session.list_keys.return_value = ["job-path/output/model.tar.gz"]
    _aws_session.region = job_region

    _braket_client_mock = Mock(meta=Mock(region_name=job_region))
    _aws_session.braket_client = _braket_client_mock
    return _aws_session


@pytest.fixture
def generate_get_job_response():
    def _get_job_response(**kwargs):
        response = {
            "ResponseMetadata": {
                "RequestId": "d223b1a0-ee5c-4c75-afa7-3c29d5338b62",
                "HTTPStatusCode": 200,
            },
            "algorithmSpecification": {
                "scriptModeConfig": {
                    "entryPoint": "my_file:start_here",
                    "s3Uri": "s3://amazon-braket-jobs/job-path/my_file.py",
                }
            },
            "checkpointConfig": {
                "localPath": "/opt/omega/checkpoints",
                "s3Uri": "s3://amazon-braket-jobs/job-path/checkpoints",
            },
            "createdAt": datetime.datetime(2021, 6, 28, 21, 4, 51),
            "deviceConfig": {
                "device": "arn:aws:braket:::device/qpu/rigetti/Aspen-10",
            },
            "hyperParameters": {
                "foo": "bar",
            },
            "inputDataConfig": [
                {
                    "channelName": "training_input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://amazon-braket-jobs/job-path/input",
                        }
                    },
                }
            ],
            "instanceConfig": {
                "instanceCount": 1,
                "instanceType": "ml.m5.large",
                "volumeSizeInGb": 1,
            },
            "jobArn": "arn:aws:braket:us-west-2:875981177017:job/job-test-20210628140446",
            "jobName": "job-test-20210628140446",
            "outputDataConfig": {"s3Path": "s3://amazon-braket-jobs/job-path/data"},
            "queueInfo": {"position": "1", "queue": "JOBS_QUEUE"},
            "roleArn": "arn:aws:iam::875981177017:role/AmazonBraketJobRole",
            "status": "RUNNING",
            "stoppingCondition": {"maxRuntimeInSeconds": 1200},
        }
        response.update(kwargs)

        return response

    return _get_job_response


@pytest.fixture
def generate_cancel_job_response():
    def _cancel_job_response(**kwargs):
        response = {
            "ResponseMetadata": {
                "RequestId": "857b0893-2073-4ad6-b828-744af8400dfe",
                "HTTPStatusCode": 200,
            },
            "cancellationStatus": "CANCELLING",
            "jobArn": "arn:aws:braket:us-west-2:875981177017:job/job-test-20210628140446",
        }
        response.update(kwargs)
        return response

    return _cancel_job_response


@pytest.fixture
def quantum_job_name():
    return "job-test-20210628140446"


@pytest.fixture
def job_region():
    return "us-west-2"


@pytest.fixture
def quantum_job_arn(quantum_job_name, job_region):
    return f"arn:aws:braket:{job_region}:875981177017:job/{quantum_job_name}"


@pytest.fixture
def quantum_job(quantum_job_arn, aws_session):
    return AwsQuantumJob(quantum_job_arn, aws_session)


def test_equality(quantum_job_arn, aws_session, job_region):
    new_aws_session = Mock(region=job_region)
    quantum_job_1 = AwsQuantumJob(quantum_job_arn, aws_session)
    quantum_job_2 = AwsQuantumJob(quantum_job_arn, aws_session)
    quantum_job_3 = AwsQuantumJob(quantum_job_arn, new_aws_session)
    other_quantum_job = AwsQuantumJob(
        "arn:aws:braket:us-west-2:875981177017:job/other-job", aws_session
    )
    non_quantum_job = quantum_job_1.arn

    assert quantum_job_1 == quantum_job_2
    assert quantum_job_1 == quantum_job_3
    assert quantum_job_1 is not quantum_job_2
    assert quantum_job_1 is not quantum_job_3
    assert quantum_job_1 is quantum_job_1
    assert quantum_job_1 != other_quantum_job
    assert quantum_job_1 != non_quantum_job


def test_hash(quantum_job):
    assert hash(quantum_job) == hash(quantum_job.arn)


@pytest.mark.parametrize(
    "arn, expected_region",
    [
        ("arn:aws:braket:us-west-2:875981177017:job/job-name", "us-west-2"),
        ("arn:aws:braket:us-west-1:1234567890:job/job-name", "us-west-1"),
    ],
)
@patch("braket.aws.aws_quantum_job.boto3.Session")
@patch("braket.aws.aws_quantum_job.AwsSession")
def test_quantum_job_constructor_default_session(
    aws_session_mock, mock_session, arn, expected_region
):
    mock_boto_session = Mock()
    aws_session_mock.return_value = Mock()
    mock_session.return_value = mock_boto_session
    job = AwsQuantumJob(arn)
    mock_session.assert_called_with(region_name=expected_region)
    aws_session_mock.assert_called_with(boto_session=mock_boto_session)
    assert job.arn == arn
    assert job._aws_session == aws_session_mock.return_value


def test_quantum_job_constructor_invalid_region(aws_session):
    region_mismatch = "The aws session region does not match the region for the supplied arn."
    arn = "arn:aws:braket:unknown-region:875981177017:job/quantum_job_name"
    with pytest.raises(ValueError, match=region_mismatch):
        AwsQuantumJob(arn, aws_session)


@patch("braket.aws.aws_quantum_job.boto3.Session")
def test_quantum_job_constructor_explicit_session(mock_session, quantum_job_arn, job_region):
    aws_session_mock = Mock(region=job_region)
    job = AwsQuantumJob(quantum_job_arn, aws_session_mock)
    assert job._aws_session == aws_session_mock
    assert job.arn == quantum_job_arn
    mock_session.assert_not_called()


def test_metadata(quantum_job, aws_session, generate_get_job_response, quantum_job_arn):
    get_job_response_running = generate_get_job_response(status="RUNNING")
    aws_session.get_job.return_value = get_job_response_running
    assert quantum_job.metadata() == get_job_response_running
    aws_session.get_job.assert_called_with(quantum_job_arn)

    get_job_response_completed = generate_get_job_response(status="COMPLETED")
    aws_session.get_job.return_value = get_job_response_completed
    assert quantum_job.metadata() == get_job_response_completed
    aws_session.get_job.assert_called_with(quantum_job_arn)
    assert aws_session.get_job.call_count == 2


def test_metadata_caching(quantum_job, aws_session, generate_get_job_response, quantum_job_arn):
    get_job_response_running = generate_get_job_response(status="RUNNING")
    aws_session.get_job.return_value = get_job_response_running
    assert quantum_job.metadata(True) == get_job_response_running

    get_job_response_completed = generate_get_job_response(status="COMPLETED")
    aws_session.get_job.return_value = get_job_response_completed
    assert quantum_job.metadata(True) == get_job_response_running
    aws_session.get_job.assert_called_with(quantum_job_arn)
    assert aws_session.get_job.call_count == 1


def test_queue_position(quantum_job, aws_session, generate_get_job_response):
    state_1 = "COMPLETED"
    queue_info = {
        "queue": "JOBS_QUEUE",
        "position": "None",
        "message": "Job is in COMPLETED status. "
        "AmazonBraket does not show queue position for this status.",
    }
    get_job_response_completed = generate_get_job_response(status=state_1, queueInfo=queue_info)
    aws_session.get_job.return_value = get_job_response_completed
    assert quantum_job.queue_position() == HybridJobQueueInfo(
        queue_position=None, message=queue_info["message"]
    )

    state_2 = "QUEUED"
    queue_info = {"queue": "JOBS_QUEUE", "position": "2"}
    get_job_response_queued = generate_get_job_response(status=state_2, queueInfo=queue_info)
    aws_session.get_job.return_value = get_job_response_queued
    assert quantum_job.queue_position() == HybridJobQueueInfo(queue_position="2", message=None)


def test_state(quantum_job, aws_session, generate_get_job_response, quantum_job_arn):
    state_1 = "RUNNING"
    get_job_response_running = generate_get_job_response(status=state_1)
    aws_session.get_job.return_value = get_job_response_running
    assert quantum_job.state() == state_1
    aws_session.get_job.assert_called_with(quantum_job_arn)

    state_2 = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state_2)
    aws_session.get_job.return_value = get_job_response_completed
    assert quantum_job.state() == state_2
    aws_session.get_job.assert_called_with(quantum_job_arn)
    assert aws_session.get_job.call_count == 2


def test_state_caching(quantum_job, aws_session, generate_get_job_response, quantum_job_arn):
    state_1 = "RUNNING"
    get_job_response_running = generate_get_job_response(status=state_1)
    aws_session.get_job.return_value = get_job_response_running
    assert quantum_job.state(True) == state_1

    state_2 = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state_2)
    aws_session.get_job.return_value = get_job_response_completed
    assert quantum_job.state(True) == state_1
    aws_session.get_job.assert_called_with(quantum_job_arn)
    assert aws_session.get_job.call_count == 1


@pytest.fixture()
def result_setup(quantum_job_name):
    with tempfile.TemporaryDirectory() as temp_dir:
        os.chdir(temp_dir)
        file_path = "results.json"

        with open(file_path, "w") as write_file:
            write_file.write(
                json.dumps(
                    {
                        "braketSchemaHeader": {
                            "name": "braket.jobs_data.persisted_job_data",
                            "version": "1",
                        },
                        "dataDictionary": {"converged": True, "energy": -0.2},
                        "dataFormat": "plaintext",
                    }
                )
            )

        with tarfile.open("model.tar.gz", "w:gz") as tar:
            tar.add(file_path, arcname=os.path.basename(file_path))

        yield

        result_dir = f"{os.getcwd()}/{quantum_job_name}"

        if os.path.exists(result_dir):
            os.remove(f"{result_dir}/results.json")
            os.rmdir(f"{result_dir}/")

        if os.path.isfile("model.tar.gz"):
            os.remove("model.tar.gz")

        os.chdir("..")


@pytest.mark.parametrize("state", sorted(AwsQuantumJob.TERMINAL_STATES))
def test_results_when_job_is_completed(
    quantum_job, aws_session, generate_get_job_response, result_setup, state
):
    expected_saved_data = {"converged": True, "energy": -0.2}

    get_job_response_completed = generate_get_job_response(status=state)
    quantum_job._aws_session.get_job.return_value = get_job_response_completed
    actual_data = quantum_job.result()

    job_metadata = quantum_job.metadata(True)
    s3_path = job_metadata["outputDataConfig"]["s3Path"]

    output_bucket_uri = f"{s3_path}/output/model.tar.gz"
    quantum_job._aws_session.download_from_s3.assert_called_with(
        s3_uri=output_bucket_uri, filename="model.tar.gz"
    )
    assert actual_data == expected_saved_data


def test_download_result_when_job_is_running(
    quantum_job, aws_session, generate_get_job_response, result_setup
):
    poll_timeout_seconds, poll_interval_seconds, state = 1, 0.5, "RUNNING"
    get_job_response_completed = generate_get_job_response(status=state)
    aws_session.get_job.return_value = get_job_response_completed
    job_metadata = quantum_job.metadata(True)

    with pytest.raises(
        TimeoutError,
        match=f"{job_metadata['jobName']}: Polling for job completion "
        f"timed out after {poll_timeout_seconds} seconds.",
    ):
        quantum_job.download_result(
            poll_timeout_seconds=poll_timeout_seconds, poll_interval_seconds=poll_interval_seconds
        )


def test_download_result_when_extract_path_not_provided(
    quantum_job, generate_get_job_response, aws_session, result_setup
):
    state = "COMPLETED"
    expected_saved_data = {"converged": True, "energy": -0.2}
    get_job_response_completed = generate_get_job_response(status=state)
    quantum_job._aws_session.get_job.return_value = get_job_response_completed
    job_metadata = quantum_job.metadata(True)
    job_name = job_metadata["jobName"]
    quantum_job.download_result()

    with open(f"{job_name}/results.json", "r") as file:
        actual_data = json.loads(file.read())["dataDictionary"]
        assert expected_saved_data == actual_data


def test_download_result_when_extract_path_provided(
    quantum_job, generate_get_job_response, aws_session, result_setup
):
    expected_saved_data = {"converged": True, "energy": -0.2}
    state = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state)
    aws_session.get_job.return_value = get_job_response_completed
    job_metadata = quantum_job.metadata(True)
    job_name = job_metadata["jobName"]

    with tempfile.TemporaryDirectory() as temp_dir:
        quantum_job.download_result(temp_dir)

        with open(f"{temp_dir}/{job_name}/results.json", "r") as file:
            actual_data = json.loads(file.read())["dataDictionary"]
            assert expected_saved_data == actual_data


def test_empty_dict_returned_when_result_not_saved(
    quantum_job, generate_get_job_response, aws_session
):
    state = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state)
    aws_session.get_job.return_value = get_job_response_completed

    exception_response = {
        "Error": {
            "Code": "404",
            "Message": "Not Found",
        }
    }
    quantum_job._aws_session.download_from_s3 = Mock(
        side_effect=ClientError(exception_response, "HeadObject")
    )
    assert quantum_job.result() == {}


def test_results_not_in_s3_for_download(quantum_job, generate_get_job_response, aws_session):
    state = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state)
    aws_session.get_job.return_value = get_job_response_completed
    job_metadata = quantum_job.metadata(True)
    output_s3_path = job_metadata["outputDataConfig"]["s3Path"]

    error_message = f"Error retrieving results, could not find results at '{output_s3_path}"

    exception_response = {
        "Error": {
            "Code": "404",
            "Message": "Not Found",
        }
    }
    quantum_job._aws_session.download_from_s3 = Mock(
        side_effect=ClientError(exception_response, "HeadObject")
    )
    with pytest.raises(ClientError, match=error_message):
        quantum_job.download_result()


def test_results_raises_error_for_non_404_errors(
    quantum_job, generate_get_job_response, aws_session
):
    state = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state)
    aws_session.get_job.return_value = get_job_response_completed

    error = "An error occurred \\(402\\) when calling the SomeObject operation: Something"

    exception_response = {
        "Error": {
            "Code": "402",
            "Message": "Something",
        }
    }
    quantum_job._aws_session.download_from_s3 = Mock(
        side_effect=ClientError(exception_response, "SomeObject")
    )
    with pytest.raises(ClientError, match=error):
        quantum_job.result()


@patch("braket.aws.aws_quantum_job.AwsQuantumJob.download_result")
def test_results_json_file_not_in_tar(
    result_download, quantum_job, aws_session, generate_get_job_response
):
    state = "COMPLETED"
    get_job_response_completed = generate_get_job_response(status=state)
    quantum_job._aws_session.get_job.return_value = get_job_response_completed
    assert quantum_job.result() == {}


@pytest.fixture
def entry_point():
    return "test-source-module.entry_point:func"


@pytest.fixture
def bucket():
    return "braket-region-id"


@pytest.fixture(
    params=[
        None,
        "aws.location/custom-jobs:tag.1.2.3",
        "other.uri/custom-name:tag",
        "other-custom-format.com",
    ]
)
def image_uri(request):
    return request.param


@pytest.fixture(params=["given_job_name", "default_job_name"])
def job_name(request):
    if request.param == "given_job_name":
        return "test-job-name"


@pytest.fixture
def s3_prefix(job_name):
    return f"{job_name}/non-default"


@pytest.fixture(params=["local_source", "s3_source"])
def source_module(request, bucket, s3_prefix):
    if request.param == "local_source":
        return "test-source-module"
    elif request.param == "s3_source":
        return AwsSession.construct_s3_uri(bucket, "test-source-prefix", "source.tar.gz")


@pytest.fixture
def role_arn():
    return "arn:aws:iam::0000000000:role/AmazonBraketInternalSLR"


@pytest.fixture(
    params=[
        "arn:aws:braket:us-west-2::device/qpu/test/device-name",
        "arn:aws:braket:::device/qpu/test/device-name",
    ]
)
def device_arn(request):
    return request.param


@pytest.fixture
def reservation_arn():
    return "arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"


@pytest.fixture
def prepare_job_args(aws_session, device_arn, reservation_arn):
    return {
        "device": device_arn,
        "source_module": Mock(),
        "entry_point": Mock(),
        "image_uri": Mock(),
        "job_name": Mock(),
        "code_location": Mock(),
        "role_arn": Mock(),
        "hyperparameters": Mock(),
        "input_data": Mock(),
        "instance_config": Mock(),
        "distribution": Mock(),
        "stopping_condition": Mock(),
        "output_data_config": Mock(),
        "copy_checkpoints_from_job": Mock(),
        "checkpoint_config": Mock(),
        "aws_session": aws_session,
        "tags": Mock(),
        "reservation_arn": reservation_arn,
    }


def test_str(quantum_job):
    expected = f"AwsQuantumJob('arn':'{quantum_job.arn}')"
    assert str(quantum_job) == expected


def test_arn(quantum_job_arn, aws_session):
    quantum_job = AwsQuantumJob(quantum_job_arn, aws_session)
    assert quantum_job.arn == quantum_job_arn


def test_name(quantum_job_arn, quantum_job_name, aws_session, generate_get_job_response):
    quantum_job = AwsQuantumJob(quantum_job_arn, aws_session)
    aws_session.get_job.return_value = generate_get_job_response(jobName=quantum_job_name)
    assert quantum_job.name == quantum_job_name


def test_no_arn_setter(quantum_job):
    # Python 3.11 error output differs from Python 3.10 <=
    with pytest.raises(AttributeError):
        quantum_job.arn = 123


@pytest.mark.parametrize("wait_until_complete", [True, False])
@patch("braket.aws.aws_quantum_job.AwsQuantumJob.logs")
@patch("braket.aws.aws_quantum_job.prepare_quantum_job")
def test_create_job(
    mock_prepare_quantum_job,
    mock_logs,
    aws_session,
    prepare_job_args,
    quantum_job_arn,
    wait_until_complete,
):
    test_response_args = {"testArgs": "MyTestArg"}
    mock_prepare_quantum_job.return_value = test_response_args
    job = AwsQuantumJob.create(wait_until_complete=wait_until_complete, **prepare_job_args)
    mock_prepare_quantum_job.assert_called_with(**prepare_job_args)
    aws_session.create_job.assert_called_with(**test_response_args)
    if wait_until_complete:
        mock_logs.assert_called_once()
    else:
        mock_logs.assert_not_called()
    assert job.arn == quantum_job_arn


def test_create_fake_arg():
    unexpected_kwarg = "create\\(\\) got an unexpected keyword argument 'fake_arg'"
    with pytest.raises(TypeError, match=unexpected_kwarg):
        AwsQuantumJob.create(
            device="device",
            source_module="source",
            fake_arg="fake_value",
        )


def test_cancel_job(quantum_job_arn, aws_session, generate_cancel_job_response):
    cancellation_status = "CANCELLING"
    aws_session.cancel_job.return_value = generate_cancel_job_response(
        cancellationStatus=cancellation_status
    )
    quantum_job = AwsQuantumJob(quantum_job_arn, aws_session)
    status = quantum_job.cancel()
    aws_session.cancel_job.assert_called_with(quantum_job_arn)
    assert status == cancellation_status


def test_cancel_job_surfaces_exception(quantum_job, aws_session):
    exception_response = {
        "Error": {
            "Code": "ValidationException",
            "Message": "unit-test-error",
        }
    }
    error_string = re.escape(
        "An error occurred (ValidationException) when calling the "
        "cancel_job operation: unit-test-error"
    )
    aws_session.cancel_job.side_effect = ClientError(exception_response, "cancel_job")
    with pytest.raises(ClientError, match=error_string):
        quantum_job.cancel()


@pytest.mark.parametrize(
    "generate_get_job_response_kwargs",
    [
        {
            "status": "RUNNING",
        },
        {
            "status": "COMPLETED",
        },
        {
            "status": "COMPLETED",
            "startedAt": datetime.datetime(2021, 1, 1, 1, 0, 0, 0),
        },
        {"status": "COMPLETED", "endedAt": datetime.datetime(2021, 1, 1, 1, 0, 0, 0)},
        {
            "status": "COMPLETED",
            "startedAt": datetime.datetime(2021, 1, 1, 1, 0, 0, 0),
            "endedAt": datetime.datetime(2021, 1, 1, 1, 0, 0, 0),
        },
    ],
)
@patch(
    "braket.jobs.metrics_data.cwl_insights_metrics_fetcher."
    "CwlInsightsMetricsFetcher.get_metrics_for_job"
)
def test_metrics(
    metrics_fetcher_mock,
    quantum_job,
    aws_session,
    generate_get_job_response,
    generate_get_job_response_kwargs,
):
    get_job_response_running = generate_get_job_response(**generate_get_job_response_kwargs)
    aws_session.get_job.return_value = get_job_response_running

    expected_metrics = {"Test": [1]}
    metrics_fetcher_mock.return_value = expected_metrics
    metrics = quantum_job.metrics()
    assert metrics == expected_metrics


@pytest.fixture
def log_stream_responses():
    return (
        ClientError(
            {
                "Error": {
                    "Code": "ResourceNotFoundException",
                    "Message": "This shouldn't get raised...",
                }
            },
            "DescribeLogStreams",
        ),
        {"logStreams": []},
        {"logStreams": [{"logStreamName": "stream-1"}]},
    )


@pytest.fixture
def log_events_responses():
    return (
        {"nextForwardToken": None, "events": [{"timestamp": 1, "message": "hi there #1"}]},
        {"nextForwardToken": None, "events": []},
        {
            "nextForwardToken": None,
            "events": [
                {"timestamp": 1, "message": "hi there #1"},
                {"timestamp": 2, "message": "hi there #2"},
            ],
        },
        {"nextForwardToken": None, "events": []},
        {
            "nextForwardToken": None,
            "events": [
                {"timestamp": 2, "message": "hi there #2"},
                {"timestamp": 2, "message": "hi there #2a"},
                {"timestamp": 3, "message": "hi there #3"},
            ],
        },
        {"nextForwardToken": None, "events": []},
    )


def test_logs(
    quantum_job,
    generate_get_job_response,
    log_events_responses,
    log_stream_responses,
    capsys,
):
    quantum_job._aws_session.get_job.side_effect = (
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
    )
    quantum_job._aws_session.describe_log_streams.side_effect = log_stream_responses
    quantum_job._aws_session.get_log_events.side_effect = log_events_responses

    quantum_job.logs(wait=True, poll_interval_seconds=0)

    captured = capsys.readouterr()
    assert captured.out == "\n".join(
        (
            "..",
            "hi there #1",
            "hi there #2",
            "hi there #2a",
            "hi there #3",
            "",
        )
    )


def test_logs_queue_progress(
    quantum_job,
    generate_get_job_response,
    log_events_responses,
    log_stream_responses,
    capsys,
):
    queue_info = {"queue": "JOBS_QUEUE", "position": "1"}
    quantum_job._aws_session.get_job.side_effect = (
        generate_get_job_response(status="QUEUED", queue_info=queue_info),
        generate_get_job_response(status="QUEUED", queue_info=queue_info),
        generate_get_job_response(status="QUEUED", queue_info=queue_info),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
    )
    quantum_job._aws_session.describe_log_streams.side_effect = log_stream_responses
    quantum_job._aws_session.get_log_events.side_effect = log_events_responses

    quantum_job.logs(wait=True, poll_interval_seconds=0)

    captured = capsys.readouterr()
    assert captured.out == "\n".join(
        (
            f"Job queue position: {queue_info['position']}",
            "Running:",
            "",
            "hi there #1",
            "hi there #2",
            "hi there #2a",
            "hi there #3",
            "",
        )
    )


@patch.dict("os.environ", {"JPY_PARENT_PID": "True"})
def test_logs_multiple_instances(
    quantum_job,
    generate_get_job_response,
    log_events_responses,
    log_stream_responses,
    capsys,
):
    quantum_job._aws_session.get_job.side_effect = (
        generate_get_job_response(status="RUNNING", instanceConfig={"instanceCount": 2}),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
        generate_get_job_response(status="COMPLETED"),
    )
    log_stream_responses[-1]["logStreams"].append({"logStreamName": "stream-2"})
    quantum_job._aws_session.describe_log_streams.side_effect = log_stream_responses

    event_counts = {
        "stream-1": 0,
        "stream-2": 0,
    }

    def get_log_events(log_group, log_stream, start_time, start_from_head, next_token):
        log_events_dict = {
            "stream-1": log_events_responses,
            "stream-2": log_events_responses,
        }
        log_events_dict["stream-1"] += (
            {
                "nextForwardToken": None,
                "events": [],
            },
            {
                "nextForwardToken": None,
                "events": [],
            },
        )
        log_events_dict["stream-2"] += (
            {
                "nextForwardToken": None,
                "events": [
                    {"timestamp": 3, "message": "hi there #3"},
                    {"timestamp": 4, "message": "hi there #4"},
                ],
            },
            {
                "nextForwardToken": None,
                "events": [],
            },
        )
        event_counts[log_stream] += 1
        return log_events_dict[log_stream][event_counts[log_stream]]

    quantum_job._aws_session.get_log_events.side_effect = get_log_events

    quantum_job.logs(wait=True, poll_interval_seconds=0)

    captured = capsys.readouterr()
    assert captured.out == "\n".join(
        (
            "..",
            "\x1b[34mhi there #1\x1b[0m",
            "\x1b[35mhi there #1\x1b[0m",
            "\x1b[34mhi there #2\x1b[0m",
            "\x1b[35mhi there #2\x1b[0m",
            "\x1b[34mhi there #2a\x1b[0m",
            "\x1b[35mhi there #2a\x1b[0m",
            "\x1b[34mhi there #3\x1b[0m",
            "\x1b[35mhi there #3\x1b[0m",
            "\x1b[35mhi there #4\x1b[0m",
            "",
        )
    )


def test_logs_error(quantum_job, generate_get_job_response, capsys):
    quantum_job._aws_session.get_job.side_effect = (
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="RUNNING"),
        generate_get_job_response(status="COMPLETED"),
    )
    quantum_job._aws_session.describe_log_streams.side_effect = (
        ClientError(
            {
                "Error": {
                    "Code": "UnknownCode",
                    "Message": "Some error message",
                }
            },
            "DescribeLogStreams",
        ),
    )

    with pytest.raises(ClientError, match="Some error message"):
        quantum_job.logs(wait=True, poll_interval_seconds=0)


def test_initialize_session_for_valid_non_regional_device(aws_session, caplog):
    device_arn = "arn:aws:braket:::device/qpu/test/device-name"
    first_region = aws_session.region
    logger = logging.getLogger(__name__)

    aws_session.get_device.side_effect = [
        ClientError(
            {
                "Error": {
                    "Code": "ResourceNotFoundException",
                }
            },
            "getDevice",
        ),
        ClientError(
            {
                "Error": {
                    "Code": "ResourceNotFoundException",
                }
            },
            "getDevice",
        ),
        device_arn,
    ]

    caplog.set_level(logging.INFO)
    AwsQuantumJob._initialize_session(aws_session, device_arn, logger)

    assert f"Changed session region from '{first_region}' to '{aws_session.region}'" in caplog.text


def test_initialize_session_for_valid_regional_device(aws_session, caplog):
    device_arn = f"arn:aws:braket:{aws_session.region}::device/qpu/test/device-name"
    logger = logging.getLogger(__name__)
    aws_session.get_device.return_value = device_arn
    caplog.set_level(logging.INFO)
    AwsQuantumJob._initialize_session(aws_session, device_arn, logger)
    assert not caplog.text


@pytest.mark.parametrize(
    "get_device_side_effect, expected_exception",
    [
        (
            [
                ClientError(
                    {
                        "Error": {
                            "Code": "ResourceNotFoundException",
                        }
                    },
                    "getDevice",
                )
            ],
            ValueError,
        ),
        (
            [
                ClientError(
                    {
                        "Error": {
                            "Code": "ThrottlingException",
                        }
                    },
                    "getDevice",
                )
            ],
            ClientError,
        ),
    ],
)
def test_regional_device_raises_error(
    get_device_side_effect, expected_exception, aws_session, caplog
):
    device_arn = f"arn:aws:braket:{aws_session.region}::device/qpu/test/device-name"
    aws_session.get_device.side_effect = get_device_side_effect
    logger = logging.getLogger(__name__)
    caplog.set_level(logging.INFO)
    with pytest.raises(expected_exception):
        AwsQuantumJob._initialize_session(aws_session, device_arn, logger)
        aws_session.get_device.assert_called_with(device_arn)
        assert not caplog.text


def test_regional_device_switches(aws_session, caplog):
    original_region = aws_session.region
    device_region = "us-east-1"
    device_arn = f"arn:aws:braket:{device_region}::device/qpu/test/device-name"
    mock_session = Mock()
    mock_session.get_device.side_effect = device_arn
    aws_session.copy_session.side_effect = [mock_session]
    logger = logging.getLogger(__name__)
    caplog.set_level(logging.INFO)

    assert mock_session == AwsQuantumJob._initialize_session(aws_session, device_arn, logger)

    aws_session.copy_session.assert_called_with(region=device_region)
    mock_session.get_device.assert_called_with(device_arn)
    assert f"Changed session region from '{original_region}' to '{device_region}'" in caplog.text


def test_initialize_session_for_invalid_device(aws_session, device_arn):
    logger = logging.getLogger(__name__)
    aws_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
            }
        },
        "getDevice",
    )

    device_not_found = f"'{device_arn}' not found."
    with pytest.raises(ValueError, match=device_not_found):
        AwsQuantumJob._initialize_session(aws_session, device_arn, logger)


def test_no_region_routing_simulator(aws_session):
    logger = logging.getLogger(__name__)

    aws_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
            }
        },
        "getDevice",
    )

    device_arn = "arn:aws:braket:::device/simulator/test/device-name"
    device_not_found = f"Simulator '{device_arn}' not found in 'us-west-2'"
    with pytest.raises(ValueError, match=device_not_found):
        AwsQuantumJob._initialize_session(aws_session, device_arn, logger)


def test_exception_in_credentials_session_region(device_arn, aws_session):
    logger = logging.getLogger(__name__)

    aws_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "SomeOtherErrorMessage",
            }
        },
        "getDevice",
    )

    error_message = (
        "An error occurred \\(SomeOtherErrorMessage\\) "
        "when calling the getDevice operation: Unknown"
    )
    with pytest.raises(ClientError, match=error_message):
        AwsQuantumJob._initialize_session(aws_session, device_arn, logger)


def test_exceptions_in_all_device_regions(aws_session):
    device_arn = "arn:aws:braket:::device/qpu/test/device-name"
    logger = logging.getLogger(__name__)

    aws_session.get_device.side_effect = [
        ClientError(
            {
                "Error": {
                    "Code": "ResourceNotFoundException",
                }
            },
            "getDevice",
        ),
        ClientError(
            {
                "Error": {
                    "Code": "SomeOtherErrorMessage",
                }
            },
            "getDevice",
        ),
    ]

    error_message = (
        "An error occurred \\(SomeOtherErrorMessage\\) "
        "when calling the getDevice operation: Unknown"
    )
    with pytest.raises(ClientError, match=error_message):
        AwsQuantumJob._initialize_session(aws_session, device_arn, logger)


@patch("braket.aws.aws_quantum_job.AwsSession")
def test_initialize_session_local_device(mock_new_session, aws_session):
    logger = logging.getLogger(__name__)
    device = "local:provider.device.name"
    # don't change a provided AwsSession
    assert AwsQuantumJob._initialize_session(aws_session, device, logger) == aws_session
    # otherwise, create an AwsSession with the profile defaults
    assert AwsQuantumJob._initialize_session(None, device, logger) == mock_new_session()


def test_bad_device_arn_format(aws_session):
    logger = logging.getLogger(__name__)
    device_not_found = (
        "Device ARN is not a valid format: bad-arn-format. For valid Braket ARNs, "
        "see 'https://docs.aws.amazon.com/braket/latest/developerguide/braket-devices.html'"
    )

    with pytest.raises(ValueError, match=device_not_found):
        AwsQuantumJob._initialize_session(aws_session, "bad-arn-format", logger)


def test_logs_prefix(job_region, quantum_job_name, aws_session, generate_get_job_response):
    aws_session.get_job.return_value = generate_get_job_response(jobName=quantum_job_name)

    # old jobs with the `arn:.../job-name` style ARN use `job-name/` as the logs prefix
    name_arn = f"arn:aws:braket:{job_region}:875981177017:job/{quantum_job_name}"
    quantum_job = AwsQuantumJob(name_arn, aws_session)
    assert quantum_job._logs_prefix == f"{quantum_job_name}"

    # jobs with the `arn:.../uuid` style ARN use `job-name/uuid/` as the logs prefix
    uuid_1 = "UUID-123456789"
    uuid_2 = "UUID-987654321"
    uuid_arn_1 = f"arn:aws:braket:{job_region}:875981177017:job/{uuid_1}"
    uuid_job_1 = AwsQuantumJob(uuid_arn_1, aws_session)
    uuid_arn_2 = f"arn:aws:braket:{job_region}:875981177017:job/{uuid_2}"
    uuid_job_2 = AwsQuantumJob(uuid_arn_2, aws_session)
    assert (
        uuid_job_1._logs_prefix
        == f"{quantum_job_name}/{uuid_1}"
        != uuid_job_2._logs_prefix
        == f"{quantum_job_name}/{uuid_2}"
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import ast
import operator
from functools import reduce
from numbers import Number
from typing import Any, Union

import sympy
from oqpy.base import OQPyExpression
from oqpy.classical_types import FloatVar


class FreeParameterExpression:
    """Class 'FreeParameterExpression'

    Objects that can take a parameter all inherit from :class:'Parameterizable'.
    FreeParametersExpressions can hold FreeParameters that can later be
    swapped out for a number. Circuits or PulseSequences with FreeParameters
    present will NOT run. Values must be substituted prior to execution.
    """

    def __init__(self, expression: Union[FreeParameterExpression, Number, sympy.Expr, str]):
        """Initializes a FreeParameterExpression. Best practice is to initialize using
        FreeParameters and Numbers. Not meant to be initialized directly.

        Below are examples of how FreeParameterExpressions should be made.

        Args:
            expression (Union[FreeParameterExpression, Number, Expr, str]): The expression to use.

        Raises:
            NotImplementedError: Raised if the expression is not of type
                [FreeParameterExpression, Number, Expr, str]

        Examples:
            >>> expression_1 = FreeParameter("theta") * FreeParameter("alpha")
            >>> expression_2 = 1 + FreeParameter("beta") + 2 * FreeParameter("alpha")
        """
        self._operations = {
            ast.Add: self.__add__,
            ast.Sub: self.__sub__,
            ast.Mult: self.__mul__,
            ast.Pow: self.__pow__,
            ast.USub: self.__neg__,
        }
        if isinstance(expression, FreeParameterExpression):
            self._expression = expression.expression
        elif isinstance(expression, (Number, sympy.Expr)):
            self._expression = expression
        elif isinstance(expression, str):
            self._expression = self._parse_string_expression(expression).expression
        else:
            raise NotImplementedError

    @property
    def expression(self) -> Union[Number, sympy.Expr]:
        """Gets the expression.

        Returns:
            Union[Number, Expr]: The expression for the FreeParameterExpression.
        """
        return self._expression

    def subs(
        self, parameter_values: dict[str, Number]
    ) -> Union[FreeParameterExpression, Number, sympy.Expr]:
        """
        Similar to a substitution in Sympy. Parameters are swapped for corresponding values or
        expressions from the dictionary.

        Args:
            parameter_values (dict[str, Number]): A mapping of parameters to their corresponding
                values to be assigned.

        Returns:
            Union[FreeParameterExpression, Number, Expr]: A numerical value if there are no
            symbols left in the expression otherwise returns a new FreeParameterExpression.
        """
        new_parameter_values = {}
        for key, val in parameter_values.items():
            if issubclass(type(key), FreeParameterExpression):
                new_parameter_values[key.expression] = val
            else:
                new_parameter_values[key] = val

        subbed_expr = self._expression.subs(new_parameter_values)
        if isinstance(subbed_expr, Number):
            return subbed_expr
        else:
            return FreeParameterExpression(subbed_expr)

    def _parse_string_expression(self, expression: str) -> FreeParameterExpression:
        return self._eval_operation(ast.parse(expression, mode="eval").body)

    def _eval_operation(self, node: Any) -> FreeParameterExpression:
        if isinstance(node, ast.Constant):
            return FreeParameterExpression(node.n)
        elif isinstance(node, ast.Name):
            return FreeParameterExpression(sympy.Symbol(node.id))
        elif isinstance(node, ast.BinOp):
            if type(node.op) not in self._operations.keys():
                raise ValueError(f"Unsupported binary operation: {type(node.op)}")
            return self._eval_operation(node.left)._operations[type(node.op)](
                self._eval_operation(node.right)
            )
        elif isinstance(node, ast.UnaryOp):
            if type(node.op) not in self._operations.keys():
                raise ValueError(f"Unsupported unary operation: {type(node.op)}", type(node.op))
            return self._eval_operation(node.operand)._operations[type(node.op)]()
        else:
            raise ValueError(f"Unsupported string detected: {node}")

    def __add__(self, other: FreeParameterExpression):
        if issubclass(type(other), FreeParameterExpression):
            return FreeParameterExpression(self.expression + other.expression)
        else:
            return FreeParameterExpression(self.expression + other)

    def __radd__(self, other: FreeParameterExpression):
        return FreeParameterExpression(other + self.expression)

    def __sub__(self, other: FreeParameterExpression):
        if issubclass(type(other), FreeParameterExpression):
            return FreeParameterExpression(self.expression - other.expression)
        else:
            return FreeParameterExpression(self.expression - other)

    def __rsub__(self, other: FreeParameterExpression):
        return FreeParameterExpression(other - self.expression)

    def __mul__(self, other: FreeParameterExpression):
        if issubclass(type(other), FreeParameterExpression):
            return FreeParameterExpression(self.expression * other.expression)
        else:
            return FreeParameterExpression(self.expression * other)

    def __rmul__(self, other: FreeParameterExpression):
        return FreeParameterExpression(other * self.expression)

    def __truediv__(self, other):
        if issubclass(type(other), FreeParameterExpression):
            return FreeParameterExpression(self.expression / other.expression)
        else:
            return FreeParameterExpression(self.expression / other)

    def __rtruediv__(self, other: FreeParameterExpression):
        return FreeParameterExpression(other / self.expression)

    def __pow__(self, other: FreeParameterExpression, modulo: float = None):
        if issubclass(type(other), FreeParameterExpression):
            return FreeParameterExpression(self.expression**other.expression)
        else:
            return FreeParameterExpression(self.expression**other)

    def __rpow__(self, other: FreeParameterExpression):
        return FreeParameterExpression(other**self.expression)

    def __neg__(self):
        return FreeParameterExpression(-1 * self.expression)

    def __eq__(self, other: FreeParameterExpression):
        if isinstance(other, FreeParameterExpression):
            return sympy.sympify(self.expression).equals(sympy.sympify(other.expression))
        return False

    def __repr__(self) -> str:
        """The representation of the :class:'FreeParameterExpression'.

        Returns:
            str: The expression of the class:'FreeParameterExpression' to represent the class.
        """
        return repr(self.expression)

    def _to_oqpy_expression(self) -> OQPyExpression:
        """Transforms into an OQPyExpression.

        Returns:
            OQPyExpression: The AST node.
        """
        ops = {sympy.Add: operator.add, sympy.Mul: operator.mul, sympy.Pow: operator.pow}
        if isinstance(self.expression, tuple(ops)):
            return reduce(
                ops[type(self.expression)],
                map(
                    lambda x: FreeParameterExpression(x)._to_oqpy_expression(), self.expression.args
                ),
            )
        elif isinstance(self.expression, sympy.Number):
            return float(self.expression)
        else:
            fvar = FloatVar(
                name=self.expression.name, init_expression="input", needs_declaration=False
            )
            fvar.size = None
            fvar.type.size = None
            return fvar


def subs_if_free_parameter(parameter: Any, **kwargs: Union[FreeParameterExpression, str]) -> Any:
    """Substitute a free parameter with the given kwargs, if any.

    Args:
        parameter (Any): The parameter.
        **kwargs (Union[FreeParameterExpression, str]): The kwargs to use to substitute.

    Returns:
        Any: The substituted parameters.
    """
    if isinstance(parameter, FreeParameterExpression):
        substituted = parameter.subs(kwargs)
        if isinstance(substituted, sympy.Number):
            substituted = float(substituted)
        return substituted
    return parameter


def _is_float(argument: str) -> bool:
    """Checks if a string can be cast into a float.

    Args:
        argument (str): String to check.

    Returns:
        bool: Returns true if the string can be cast as a float. False, otherwise.

    """
    try:
        float(argument)
        return True
    except ValueError:
        return False



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
from decimal import Decimal
from unittest.mock import Mock

import numpy as np
import pytest

from braket.ahs.analog_hamiltonian_simulation import (
    AnalogHamiltonianSimulation,
    AtomArrangement,
    DiscretizationError,
    DrivingField,
    ShiftingField,
    SiteType,
)
from braket.ahs.atom_arrangement import AtomArrangementItem
from braket.ahs.field import Field
from braket.ahs.pattern import Pattern
from braket.ir.ahs.program_v1 import Program
from braket.timings.time_series import TimeSeries


@pytest.fixture
def register():
    return (
        AtomArrangement()
        .add((0.0, 0.0))
        .add((0.0, 3.0e-6))
        .add((0.0, 6.0e-6))
        .add((3.0e-6, 0.0))
        .add((3.0e-6, 3.0e-6))
        .add((3.0e-6, 3.0e-6), SiteType.VACANT)
        .add((3.0e-6, 6.0e-6), SiteType.VACANT)
    )


@pytest.fixture
def driving_field():
    return DrivingField(
        TimeSeries().put(0.0, 0.0).put(3.0e-7, 2.51327e7).put(2.7e-6, 2.51327e7).put(3.0e-6, 0.0),
        TimeSeries().put(0.0, 0).put(3.0e-6, 0),
        TimeSeries()
        .put(0.0, -1.25664e8)
        .put(3.0e-7, -1.25664e8)
        .put(2.7e-6, 1.25664e8)
        .put(3.0e-6, 1.25664e8),
    )


@pytest.fixture
def shifting_field():
    return ShiftingField(
        Field(
            TimeSeries().put(0.0, -1.25664e8).put(3.0e-6, 1.25664e8),
            Pattern([0.5, 1.0, 0.5, 0.5, 0.5, 0.5]),
        )
    )


def test_create():
    mock0 = Mock()
    mock1 = Mock()
    ahs = AnalogHamiltonianSimulation(register=mock0, hamiltonian=mock1)
    assert mock0 == ahs.register
    assert mock1 == ahs.hamiltonian


def test_to_ir(register, driving_field, shifting_field):
    hamiltonian = driving_field + shifting_field
    ahs = AnalogHamiltonianSimulation(register=register, hamiltonian=hamiltonian)
    problem = ahs.to_ir()
    assert Program.parse_raw(problem.json()) == problem
    assert problem == Program.parse_raw_schema(problem.json())


def test_to_ir_empty():
    hamiltonian = Mock()
    hamiltonian.terms = []
    ahs = AnalogHamiltonianSimulation(register=AtomArrangement(), hamiltonian=hamiltonian)
    problem = ahs.to_ir()
    assert Program.parse_raw(problem.json()) == problem
    assert problem == Program.parse_raw_schema(problem.json())


@pytest.mark.xfail(raises=TypeError)
def test_to_ir_invalid_hamiltonian(register):
    hamiltonian = Mock()
    hamiltonian.terms = [Mock()]
    ahs = AnalogHamiltonianSimulation(register=register, hamiltonian=hamiltonian)
    ahs.to_ir()


@pytest.mark.xfail(raises=DiscretizationError)
def test_invalid_action():
    action = Mock()
    action.actionType = "not-a-valid-AHS-action"
    device = Mock()
    device.properties.action = {"braket.ir.ahs.program": action}

    AnalogHamiltonianSimulation(register=Mock(), hamiltonian=Mock()).discretize(device)


@pytest.mark.xfail(raises=DiscretizationError)
def test_invalid_action_name():
    action = Mock()
    action.actionType = "braket.ir.ahs.program"
    device = Mock()
    device.properties.action = {"not-a-valid-AHS-action": action}

    AnalogHamiltonianSimulation(register=Mock(), hamiltonian=Mock()).discretize(device)


def test_discretize(register, driving_field, shifting_field):
    hamiltonian = driving_field + shifting_field
    ahs = AnalogHamiltonianSimulation(register=register, hamiltonian=hamiltonian)

    action = Mock()
    action.actionType = "braket.ir.ahs.program"

    device = Mock()
    device.properties.action = {"braket.ir.ahs.program": action}

    device.properties.paradigm.lattice.geometry.positionResolution = Decimal("1E-7")

    device.properties.paradigm.rydberg.rydbergGlobal.timeResolution = Decimal("1E-9")
    device.properties.paradigm.rydberg.rydbergGlobal.rabiFrequencyResolution = Decimal("400")
    device.properties.paradigm.rydberg.rydbergGlobal.detuningResolution = Decimal("0.2")
    device.properties.paradigm.rydberg.rydbergGlobal.phaseResolution = Decimal("5E-7")

    device.properties.paradigm.rydberg.rydbergLocal.timeResolution = Decimal("1E-9")
    device.properties.paradigm.rydberg.rydbergLocal.commonDetuningResolution = Decimal("2000.0")
    device.properties.paradigm.rydberg.rydbergLocal.localDetuningResolution = Decimal("0.01")

    discretized_ahs = ahs.discretize(device)
    discretized_ir = discretized_ahs.to_ir()
    discretized_json = json.loads(discretized_ir.json())
    assert discretized_json["setup"]["ahs_register"] == {
        "filling": [1, 1, 1, 1, 1, 0, 0],
        "sites": [
            ["0E-7", "0E-7"],
            ["0E-7", "0.0000030"],
            ["0E-7", "0.0000060"],
            ["0.0000030", "0E-7"],
            ["0.0000030", "0.0000030"],
            ["0.0000030", "0.0000030"],
            ["0.0000030", "0.0000060"],
        ],
    }
    assert discretized_json["hamiltonian"]["drivingFields"][0]["amplitude"] == {
        "pattern": "uniform",
        "time_series": {
            "times": ["0E-9", "3.00E-7", "0.000002700", "0.000003000"],
            "values": ["0", "25132800", "25132800", "0"],
        },
    }
    assert discretized_json["hamiltonian"]["drivingFields"][0]["phase"] == {
        "pattern": "uniform",
        "time_series": {"times": ["0E-9", "0.000003000"], "values": ["0E-7", "0E-7"]},
    }
    assert discretized_json["hamiltonian"]["drivingFields"][0]["detuning"] == {
        "pattern": "uniform",
        "time_series": {
            "times": ["0E-9", "3.00E-7", "0.000002700", "0.000003000"],
            "values": ["-125664000.0", "-125664000.0", "125664000.0", "125664000.0"],
        },
    }
    assert discretized_json["hamiltonian"]["shiftingFields"][0]["magnitude"] == {
        "pattern": ["0.50", "1.00", "0.50", "0.50", "0.50", "0.50"],
        "time_series": {
            "times": ["0E-9", "0.000003000"],
            "values": ["-125664000.0", "125664000.0"],
        },
    }


def test_converting_numpy_array_sites_to_ir(driving_field):
    hamiltonian = driving_field

    sites = np.array(
        [
            [0.0, 0.0],
            [0.0, 1.0e-6],
            [1e-6, 2.0e-6],
        ]
    )
    register = AtomArrangement()
    for site in sites:
        register.add(site)

    ahs = AnalogHamiltonianSimulation(register=register, hamiltonian=hamiltonian)
    sites_in_ir = ahs.to_ir().setup.ahs_register.sites
    expected_sites_in_ir = [
        [Decimal("0.0"), Decimal("0.0")],
        [Decimal("0.0"), Decimal("1e-6")],
        [Decimal("1e-6"), Decimal("2e-6")],
    ]

    assert sites_in_ir == expected_sites_in_ir


@pytest.mark.xfail(raises=ValueError)
def test_site_validation_wrong_length():
    register = AtomArrangement()
    register.add(np.array([0.0, 1e-6, -1e-6]))


@pytest.mark.xfail(raises=TypeError)
def test_site_validation_non_number():
    register = AtomArrangement()
    register.add(
        [
            "not-a-number",
            [
                "also-not-a-number",
            ],
        ]
    )


@pytest.mark.xfail(raises=TypeError)
def test_site_validation_not_a_tuple():
    AtomArrangementItem(None, SiteType.FILLED)


@pytest.mark.xfail(raises=ValueError)
def test_site_validation_invalid_site_type():
    register = AtomArrangement()
    register.add([0.0, 0.0], "not-a-valid-site-type")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from functools import singledispatchmethod
from itertools import repeat
from multiprocessing import Pool
from os import cpu_count
from typing import Any, Optional, Union

import pkg_resources

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem
from braket.circuits import Circuit
from braket.circuits.circuit_helpers import validate_circuit_and_shots
from braket.circuits.noise_model import NoiseModel
from braket.circuits.serialization import IRType
from braket.device_schema import DeviceActionType, DeviceCapabilities
from braket.devices.device import Device
from braket.ir.ahs import Program as AHSProgram
from braket.ir.openqasm import Program
from braket.simulator import BraketSimulator
from braket.tasks import AnnealingQuantumTaskResult, GateModelQuantumTaskResult
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
    AnalogHamiltonianSimulationQuantumTaskResult,
)
from braket.tasks.local_quantum_task import LocalQuantumTask
from braket.tasks.local_quantum_task_batch import LocalQuantumTaskBatch

_simulator_devices = {
    entry.name: entry for entry in pkg_resources.iter_entry_points("braket.simulators")
}


class LocalSimulator(Device):
    """A simulator meant to run directly on the user's machine.

    This class wraps a BraketSimulator object so that it can be run and returns
    results using constructs from the SDK rather than Braket IR.
    """

    def __init__(
        self,
        backend: Union[str, BraketSimulator] = "default",
        noise_model: Optional[NoiseModel] = None,
    ):
        """Initializes a `LocalSimulator`.

        Args:
            backend (Union[str, BraketSimulator]): The name of the simulator backend or
                the actual simulator instance to use for simulation. Defaults to the
                `default` simulator backend name.
            noise_model (Optional[NoiseModel]): The Braket noise model to apply to the circuit
                before execution. Noise model can only be added to the devices that support
                noise simulation.
        """
        delegate = self._get_simulator(backend)
        super().__init__(
            name=delegate.__class__.__name__,
            status="AVAILABLE",
        )
        self._delegate = delegate
        if noise_model:
            self._validate_device_noise_model_support(noise_model)
        self._noise_model = noise_model

    def run(
        self,
        task_specification: Union[Circuit, Problem, Program, AnalogHamiltonianSimulation],
        shots: int = 0,
        inputs: Optional[dict[str, float]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> LocalQuantumTask:
        """Runs the given task with the wrapped local simulator.

        Args:
            task_specification (Union[Circuit, Problem, Program, AnalogHamiltonianSimulation]):
                The quantum task specification.
            shots (int): The number of times to run the circuit or annealing problem.
                Default is 0, which means that the simulator will compute the exact
                results based on the quantum task specification.
                Sampling is not supported for shots=0.
            inputs (Optional[dict[str, float]]): Inputs to be passed along with the
                IR. If the IR supports inputs, the inputs will be updated with this
                value. Default: {}.
            *args (Any): Arbitrary arguments.
            **kwargs(Any): Arbitrary keyword arguments.

        Returns:
            LocalQuantumTask: A LocalQuantumTask object containing the results
            of the simulation

        Note:
            If running a circuit, the number of qubits will be passed
            to the backend as the argument after the circuit itself.

        Examples:
            >>> circuit = Circuit().h(0).cnot(0, 1)
            >>> device = LocalSimulator("default")
            >>> device.run(circuit, shots=1000)
        """
        if self._noise_model:
            task_specification = self._apply_noise_model_to_circuit(task_specification)
        result = self._run_internal(task_specification, shots, inputs=inputs, *args, **kwargs)
        return LocalQuantumTask(result)

    def run_batch(  # noqa: C901
        self,
        task_specifications: Union[
            Union[Circuit, Problem, Program, AnalogHamiltonianSimulation],
            list[Union[Circuit, Problem, Program, AnalogHamiltonianSimulation]],
        ],
        shots: Optional[int] = 0,
        max_parallel: Optional[int] = None,
        inputs: Optional[Union[dict[str, float], list[dict[str, float]]]] = None,
        *args,
        **kwargs,
    ) -> LocalQuantumTaskBatch:
        """Executes a batch of quantum tasks in parallel

        Args:
            task_specifications (Union[Union[Circuit, Problem, Program, AnalogHamiltonianSimulation], list[Union[Circuit, Problem, Program, AnalogHamiltonianSimulation]]]):
                Single instance or list of quantum task specification.
            shots (Optional[int]): The number of times to run the quantum task.
                Default: 0.
            max_parallel (Optional[int]): The maximum number of quantum tasks to run  in parallel. Default
                is the number of CPU.
            inputs (Optional[Union[dict[str, float], list[dict[str, float]]]]): Inputs to be passed
                along with the IR. If the IR supports inputs, the inputs will be updated with
                this value. Default: {}.

        Returns:
            LocalQuantumTaskBatch: A batch containing all of the quantum tasks run

        See Also:
            `braket.tasks.local_quantum_task_batch.LocalQuantumTaskBatch`
        """  # noqa E501
        inputs = inputs or {}

        if self._noise_model:
            task_specifications = [
                self._apply_noise_model_to_circuit(task_specification)
                for task_specification in task_specifications
            ]

        if not max_parallel:
            max_parallel = cpu_count()

        single_task = isinstance(
            task_specifications,
            (Circuit, Program, Problem, AnalogHamiltonianSimulation),
        )

        single_input = isinstance(inputs, dict)

        if not single_task and not single_input:
            if len(task_specifications) != len(inputs):
                raise ValueError("Multiple inputs and task specifications must be equal in number.")
        if single_task:
            task_specifications = repeat(task_specifications)

        if single_input:
            inputs = repeat(inputs)

        tasks_and_inputs = zip(task_specifications, inputs)

        if single_task and single_input:
            tasks_and_inputs = [next(tasks_and_inputs)]
        else:
            tasks_and_inputs = list(tasks_and_inputs)

        for task_specification, input_map in tasks_and_inputs:
            if isinstance(task_specification, Circuit):
                param_names = {param.name for param in task_specification.parameters}
                unbounded_parameters = param_names - set(input_map.keys())
                if unbounded_parameters:
                    raise ValueError(
                        f"Cannot execute circuit with unbound parameters: "
                        f"{unbounded_parameters}"
                    )

        with Pool(min(max_parallel, len(tasks_and_inputs))) as pool:
            param_list = [(task, shots, inp, *args, *kwargs) for task, inp in tasks_and_inputs]
            results = pool.starmap(self._run_internal_wrap, param_list)

        return LocalQuantumTaskBatch(results)

    @property
    def properties(self) -> DeviceCapabilities:
        """DeviceCapabilities: Return the device properties

        Please see `braket.device_schema` in amazon-braket-schemas-python_

        .. _amazon-braket-schemas-python: https://github.com/aws/amazon-braket-schemas-python
        """
        return self._delegate.properties

    @staticmethod
    def registered_backends() -> set[str]:
        """Gets the backends that have been registered as entry points

        Returns:
            set[str]: The names of the available backends that can be passed
            into LocalSimulator's constructor
        """
        return set(_simulator_devices.keys())

    def _run_internal_wrap(
        self,
        task_specification: Union[Circuit, Problem, Program, AnalogHamiltonianSimulation],
        shots: Optional[int] = None,
        inputs: Optional[dict[str, float]] = None,
        *args,
        **kwargs,
    ) -> Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult]:  # pragma: no cover
        """Wraps _run_interal for pickle dump"""
        return self._run_internal(task_specification, shots, inputs=inputs, *args, **kwargs)

    @singledispatchmethod
    def _get_simulator(self, simulator: Union[str, BraketSimulator]) -> LocalSimulator:
        raise TypeError("Simulator must either be a string or a BraketSimulator instance")

    @_get_simulator.register
    def _(self, backend_name: str):
        if backend_name in _simulator_devices:
            device_class = _simulator_devices[backend_name].load()
            return device_class()
        else:
            raise ValueError(
                f"Only the following devices are available {_simulator_devices.keys()}"
            )

    @_get_simulator.register
    def _(self, backend_impl: BraketSimulator):
        return backend_impl

    @singledispatchmethod
    def _run_internal(
        self,
        task_specification: Union[
            Circuit, Problem, Program, AnalogHamiltonianSimulation, AHSProgram
        ],
        shots: Optional[int] = None,
        *args,
        **kwargs,
    ) -> Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult]:
        raise NotImplementedError(f"Unsupported task type {type(task_specification)}")

    @_run_internal.register
    def _(
        self,
        circuit: Circuit,
        shots: Optional[int] = None,
        inputs: Optional[dict[str, float]] = None,
        *args,
        **kwargs,
    ):
        simulator = self._delegate
        if DeviceActionType.OPENQASM in simulator.properties.action:
            validate_circuit_and_shots(circuit, shots)
            program = circuit.to_ir(ir_type=IRType.OPENQASM)
            program.inputs.update(inputs or {})
            results = simulator.run(program, shots, *args, **kwargs)
            return GateModelQuantumTaskResult.from_object(results)
        elif DeviceActionType.JAQCD in simulator.properties.action:
            validate_circuit_and_shots(circuit, shots)
            program = circuit.to_ir(ir_type=IRType.JAQCD)
            qubits = circuit.qubit_count
            results = simulator.run(program, qubits, shots, *args, **kwargs)
            return GateModelQuantumTaskResult.from_object(results)
        raise NotImplementedError(f"{type(simulator)} does not support qubit gate-based programs")

    @_run_internal.register
    def _(self, problem: Problem, shots: Optional[int] = None, *args, **kwargs):
        simulator = self._delegate
        if DeviceActionType.ANNEALING not in simulator.properties.action:
            raise NotImplementedError(
                f"{type(simulator)} does not support quantum annealing problems"
            )
        ir = problem.to_ir()
        results = simulator.run(ir, shots, *args, *kwargs)
        return AnnealingQuantumTaskResult.from_object(results)

    @_run_internal.register
    def _(
        self,
        program: Program,
        shots: Optional[int] = None,
        inputs: Optional[dict[str, float]] = None,
        *args,
        **kwargs,
    ):
        simulator = self._delegate
        if DeviceActionType.OPENQASM not in simulator.properties.action:
            raise NotImplementedError(f"{type(simulator)} does not support OpenQASM programs")
        if inputs:
            inputs_copy = program.inputs.copy() if program.inputs is not None else {}
            inputs_copy.update(inputs)
            program = Program(
                source=program.source,
                inputs=inputs_copy,
            )
        results = simulator.run(program, shots, *args, **kwargs)
        return GateModelQuantumTaskResult.from_object(results)

    @_run_internal.register
    def _(
        self,
        program: AnalogHamiltonianSimulation,
        shots: Optional[int] = None,
        *args,
        **kwargs,
    ):
        simulator = self._delegate
        if DeviceActionType.AHS not in simulator.properties.action:
            raise NotImplementedError(
                f"{type(simulator)} does not support analog Hamiltonian simulation programs"
            )
        results = simulator.run(program.to_ir(), shots, *args, **kwargs)
        return AnalogHamiltonianSimulationQuantumTaskResult.from_object(results)

    @_run_internal.register
    def _(
        self,
        program: AHSProgram,
        shots: Optional[int] = None,
        *args,
        **kwargs,
    ):
        simulator = self._delegate
        if DeviceActionType.AHS not in simulator.properties.action:
            raise NotImplementedError(
                f"{type(simulator)} does not support analog Hamiltonian simulation programs"
            )
        results = simulator.run(program, shots, *args, **kwargs)
        return AnalogHamiltonianSimulationQuantumTaskResult.from_object(results)
